package edu.mit.ll.group43.surfaceoptimization;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Date;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.DefaultDepartureMinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.analysis.Logger;
import edu.mit.ll.group43.surfaceoptimization.dp.DPAlgorithmType;
import edu.mit.ll.group43.surfaceoptimization.dp.DPSequencer;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.AGOGFSpotDPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.AllGapsForceSpotDPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.AllGapsSpotDPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.DPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.LenientAllGapsForceSpotDPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.OneGapForceSpotDPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.OneGapSpotDPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.SimpleSpotDPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.input.Input;
import edu.mit.ll.group43.surfaceoptimization.input.InvalidFCFSReaderTypeException;
import edu.mit.ll.group43.surfaceoptimization.output.MatlabStochasticCodeOutput;
import edu.mit.ll.group43.surfaceoptimization.output.PlainTextOutput;
import edu.mit.ll.group43.surfaceoptimization.validator.CPSValidator;
import edu.mit.ll.group43.surfaceoptimization.validator.NoRepeatedAircraftValidator;
import edu.mit.ll.group43.surfaceoptimization.validator.ReleasedWhenReadyValidator;
import edu.mit.ll.group43.surfaceoptimization.validator.Validator;

/**
 * This is the main class of the project, which controls the various
 * sequencers, input, and output.
 * 
 * @author William Hawkins
 *
 */
public class Driver {

	// File with the schedule to optimize
	private static final String FCFS_FILE = "data/schedules/dfw20100614";
	
	// Input format
	private static final String IN_FORMAT = "text";
	
	// Which type of sequencer to use to do the optimization (fcfs, dp, or la)
	private static final String SEQ_TYPE_STR = "dp";
	
	// Which type of DP algorithm to use, if the sequencer being used is dp
	// (ss, og, ag, ogf, agogf, agf, or lagf)
	private static final String DP_ALG_STR = "agogf";
	
	// Value of the constrained position shifting variable (maximum number of
	// positions any one aircraft can shift from its original spot release
	// sequence position to its optimal spot release sequence position).
	private static final int CPS = 1;
	
	// Minimum separation calculator used
	private static DefaultDepartureMinimumSeparationCalculator calc;
	
	// Whether to collect and print meta data about the running of the code
	// (for algorithm analysis purposes)
	private static final boolean ALLOW_LOGGING = false;
	
	// Whether to collect and print running times for various steps in
	// solving the optimization problem
	private static final boolean ALLOW_TIMING = true;
	
	// Whether the MATLAB code to apply stochasticity should be output
	private static final boolean PRINT_MATLAB = false;
	
	/**
	 * Used to output important results of the algorithm, such as delays.
	 * The default out stream is System.out.
	 */
	public static PrintWriter out = new PrintWriter(System.out, true);
	
	/**
	 * Used to output meta data about the algorithm, such as how long certain
	 * operations take, how many objects of a certain type are created, or
	 * anything a Logger would keep track of and report on. The default
	 * debug stream is System.out.
	 */
	public static PrintWriter debug = new PrintWriter(System.out, true);
	
	/**
	 * Used to output error information, such as when exceptions are thrown
	 * or when a sequence returned from a sequencer is invalid. The default
	 * error stream is System.err.
	 */
	public static PrintWriter err = new PrintWriter(System.err, true);
	
	/**
	 * Use the FCFSReader to construct the FCFS list as an ArrayList of Aircraft
	 * 
	 * @return
	 * @throws FileNotFoundException 
	 * @throws InvalidFCFSReaderTypeException 
	 */
	public static ArrayList<Aircraft> generateFcfsSequence(String filename) throws FileNotFoundException, InvalidFCFSReaderTypeException {
		return (new Input()).generateFcfsSequence(IN_FORMAT, FCFS_FILE);
	}
	
	public static void outputMATLABCode(ArrayList<Aircraft> sequence) {
		(new MatlabStochasticCodeOutput()).output(System.out, sequence);
	}
	
	public static ArrayList<Aircraft> sequence(ArrayList<Aircraft> fcfs) throws FileNotFoundException, InvalidSequencerTypeException, InvalidDPAlgorithmTypeException {
		
		Date start = new Date();
		calc = DefaultDepartureMinimumSeparationCalculator.getInstance();
		Date end = new Date();
		if (ALLOW_TIMING) {
			Logger.readSeparations(end.getTime()-start.getTime());
		}
		
		// Determine which sequencer should be used
		SequencerType seq_type = null;
		try {
			seq_type = SequencerType.valueOf(SEQ_TYPE_STR.toUpperCase());
		} catch (Exception e) {
			throw new InvalidSequencerTypeException("The sequencer type " + SEQ_TYPE_STR.toUpperCase() + " is invalid");
		}
		
		DPAlgorithmType dp_type = null;
		try {
			dp_type = DPAlgorithmType.valueOf(DP_ALG_STR.toUpperCase());
		} catch (Exception e) {
			throw new InvalidDPAlgorithmTypeException("The DP algorithm type " + DP_ALG_STR.toUpperCase() + " is invalid");
		}
		
		Sequencer sequencer = null;
		switch(seq_type) {
		case FCFS:
			sequencer = new FcfsSequencer(CPS, calc);
			break;
		case LA:
			sequencer = new LookAheadSequencer(CPS, calc);
			break;
		case DP:
			DPSequencerFactory factory = null;
			switch(dp_type) {
			case SS:
				factory = new SimpleSpotDPSequencerFactory(CPS, calc);
				break;
			case OG:
				factory = new OneGapSpotDPSequencerFactory(CPS, calc);
				break;
			case AG:
				factory = new AllGapsSpotDPSequencerFactory(CPS, calc);
				break;
			case OGF:
				factory = new OneGapForceSpotDPSequencerFactory(CPS, calc);
				break;
			case AGOGF:
				factory = new AGOGFSpotDPSequencerFactory(CPS, calc);
				break;
			case AGF:
				factory = new AllGapsForceSpotDPSequencerFactory(CPS, calc);
				break;
			case LAGF:
				factory = new LenientAllGapsForceSpotDPSequencerFactory(CPS, calc);
				break;
			}
			sequencer = new DPSequencer(factory);
			break;
		}
		
		ArrayList<Aircraft> sequence = sequencer.sequence(fcfs);
		return sequence;
	}
	
	public static void measureSequence(ArrayList<Aircraft> sequence) throws FileNotFoundException {
		(new PlainTextOutput()).output(System.out, sequence);
	}
	
	public static void validateSequence(ArrayList<Aircraft> sequence) {
		Validator v2 = new ReleasedWhenReadyValidator();
		Validator v3 = new NoRepeatedAircraftValidator();
		
		if (!CPSValidator.validate(sequence, CPS)) {
			err.println("Invalid CPS Sequence");
		}
		if (!v2.validate(sequence)) {
			err.println("Invalid Spot Release Time(s)");
		}
		if (!v3.validate(sequence)) {
			err.println("Repeated Aircraft");
		}
	}
	
	public static void main(String[] args) throws FileNotFoundException, InvalidFCFSReaderTypeException, InvalidSequencerTypeException, InvalidDPAlgorithmTypeException {
		
		Logger.initialize(ALLOW_LOGGING, ALLOW_TIMING);
		
		Date start = new Date();
		ArrayList<Aircraft> fcfs = generateFcfsSequence(FCFS_FILE);
		Date end = new Date();
		if (ALLOW_TIMING) {
			Logger.generateSequence(end.getTime()-start.getTime());
		}
		
		ArrayList<Aircraft> sequence = sequence(fcfs);
		measureSequence(sequence);
		validateSequence(sequence);
		if (ALLOW_LOGGING || ALLOW_TIMING) {
			debug.println(Logger.report());
		}
		if (PRINT_MATLAB) {
			outputMATLABCode(sequence);
		}
	}
}
