package edu.mit.ll.group43.surfaceoptimization.analysis;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;

/**
 * Logs and times various events that occur in the running of the code,
 * and prints results at the end, if allowed to do so.
 * 
 * NOTE: Most of the values are not actually logged in the code but are
 * suggestions for what could be logged in the future in order to gain
 * a better understanding of how the different algorithms work on the data
 * and why they arrive at the results that they do.
 * 
 * @author William Hawkins
 *
 */
public class Logger {

	public static boolean logging = false;
	public static boolean timing = false;
	
	// Timing values
	private static long readSeparationsTime;
	private static long generateSequenceTime;
	private static long buildNetworkTime;
	private static long pruneNetworkTime;
	private static long solveNetworkTime;
	
	// Values for all DP algorithms
	private static HashMap<Integer, Integer> predNodesDist;
	
	private static HashMap<Integer, Integer> tiesDist;
	private static int ties;
	private static HashMap<Integer, Integer> tieGroupsDist;
	private static int tieGroups;
	
	// Values for Gap algorithms
	private static HashMap<Integer, Integer> gapsConsideredDist;
	private static int gapsConsidered;
	
	private static HashMap<Integer, Integer> gapsTakenDist;
	private static int gapsTaken;
	
	// Values for Gap Force algorithms
	private static HashMap<Integer, Integer> forcedAmountsDist;
	private static HashMap<Integer, Integer> forcedGapsTakenDist;
	private static int forcedGaps;
	
	// Values for LAGF
	private static int largestCPSDev;
	private static int correctInvalidToValid;
	private static int efficientInvalidToValid;
	private static int invalidsTheSame;
	private static int invalidsDifferent;
	
	public static void initialize(boolean loggingAllowed, boolean timingAllowed) {
		
		logging = loggingAllowed;
		timing = timingAllowed;
		
		if (logging) {
			ties = 0;
			tieGroups = 0;
			gapsConsidered = 0;
			gapsTaken = 0;
			forcedGaps = 0;
			largestCPSDev = 0;
			correctInvalidToValid = 0;
			efficientInvalidToValid = 0;
			invalidsTheSame = 0;
			invalidsDifferent = 0;
			
			predNodesDist = new HashMap<Integer, Integer>();
			tiesDist = new HashMap<Integer, Integer>();
			tieGroupsDist = new HashMap<Integer, Integer>();
			gapsConsideredDist = new HashMap<Integer, Integer>();
			gapsTakenDist = new HashMap<Integer, Integer>();
			forcedAmountsDist = new HashMap<Integer, Integer>();
			forcedGapsTakenDist = new HashMap<Integer, Integer>();
		}
	}
	
	public static void readSeparations(long time) {
		readSeparationsTime = time;
	}
	
	public static void generateSequence(long time) {
		generateSequenceTime = time;
	}
	
	public static void buildNetwork(long time) {
		buildNetworkTime = time;
	}
	
	public static void pruneNetwork(long time) {
		pruneNetworkTime = time;
	}
	
	public static void solveNetwork(long time) {
		solveNetworkTime = time;
	}
	
	public static boolean sameSchedule(ArrayList<Aircraft> a, ArrayList<Aircraft> b) {
		if (a == null || b == null) {
			return a == b;
		}
		if (a.size() != b.size()) {
			return false;
		}
		for (int i = 0; i < a.size(); i ++) {
			if (!(a.get(i).getFcfsIndex() == b.get(i).getFcfsIndex() &&
					a.get(i).getOptReleaseTime() == b.get(i).getOptReleaseTime())) {
				return false;
			}
		}
		return true;
	}
	
	private static void updateDist(HashMap<Integer, Integer> dist, int key) {
		if (logging) {
			int value = dist.get(key) == null ? 0 : dist.get(key);
			value ++;
			dist.put(key, value);
		}
	}
	
	public static void predNodes(int numPredNodes) {
		updateDist(predNodesDist, numPredNodes);
	}
	
	public static void tie(int stage) {
		ties ++;
		updateDist(tiesDist, stage);
	}
	
	public static void tieGroup(int groupSize) {
		tieGroups ++;
		updateDist(tieGroupsDist, groupSize);
	}
	
	public static void gapConsidered(int numGapsBack) {
		gapsConsidered ++;
		updateDist(gapsConsideredDist, numGapsBack);
	}
	
	public static void gapTaken(int numGapsBack) {
		gapsTaken ++;
		updateDist(gapsTakenDist, numGapsBack);
	}
	
	public static void forcedGapTaken(int numGapsBack, int diff) {
		forcedGaps ++;
		updateDist(forcedAmountsDist, diff);
		updateDist(forcedGapsTakenDist, numGapsBack);
	}
	
	public static void CPSDev(int cpsDev) {
		if (cpsDev > largestCPSDev) {
			largestCPSDev = cpsDev;
		}
	}
	
	public static void correctInvalidToValid() {
		correctInvalidToValid ++;
	}
	
	public static void efficientInvalidToValid() {
		efficientInvalidToValid ++;
	}
	
	public static void compareInvalids(ArrayList<Aircraft> a, ArrayList<Aircraft> b) {
		if (a == b) {
			invalidsTheSame ++;
		} else {
			invalidsDifferent ++;
		}
	}
	
	private static String scalarValue(String name, int value) {
		return name + ": " + value + "\n";
	}
	
	private static String distValue(String name, HashMap<Integer, Integer> dist) {
		ArrayList<Integer> keys = new ArrayList<Integer>(dist.keySet());
		Collections.sort(keys);
		String s = name + "\n";
		for (int key : keys) {
			s += key + ": " + dist.get(key) + "\n";
		}
		return s;
	}
	
	private static String scalarAndDistValues(String sname, int value, String dname, 
			HashMap<Integer, Integer> dist) {
		String s = scalarValue(sname, value);
		s += distValue(dname, dist);
		return s;
	}
	
	public static String report() {
		String separator = "====================\n";
		String s = separator;
		if (timing) {
			s += "Read Separations:\t\t" + readSeparationsTime + " ms\n";
			s += "Generate FCFS Sequence:\t\t" + generateSequenceTime + " ms\n";
			s += "Build Network:\t\t\t" + buildNetworkTime + " ms\n";
			s += "Prune Network:\t\t\t" + pruneNetworkTime + " ms\n";
			s += "Solve Network:\t\t\t" + solveNetworkTime + " ms\n";
			s += separator;
		}
		if (logging) {
			s += distValue("Nodes by Number of Pred Nodes", predNodesDist);
			s += separator;
			s += scalarAndDistValues("Ties", ties, "Ties by Stage", tiesDist);
			s += scalarAndDistValues("Tie Groups", tieGroups, "Tie Groups by Group Size", tieGroupsDist);
			s += separator;
			s += scalarAndDistValues("Gaps Considered", gapsConsidered, "Gaps Considered by # Gaps Back",
					gapsConsideredDist);
			s += separator;
			s += scalarAndDistValues("Gaps Taken", gapsTaken, "Gaps Taken by # Gaps Back", gapsTakenDist);
			s += separator;
			s += scalarAndDistValues("Forced Gaps Taken", forcedGaps, "Forced Gaps Taken by # Gaps Back",
					forcedGapsTakenDist);
			s += distValue("Forced Gaps Taken by Force Amount", forcedAmountsDist);
			s += separator;
			s += scalarValue("Largest Deviation from CPS", largestCPSDev);
			s += separator;
			s += scalarValue("# Correct Invalid -> Valid", correctInvalidToValid);
			s += separator;
			s += scalarValue("# Efficient Invalid -> Valid", efficientInvalidToValid);
			s += separator;
			s += scalarValue("# Times Invalids Were Same", invalidsTheSame);
			s += separator;
			s += scalarValue("# Times Invalids Different", invalidsDifferent);
		}
		
		return s;
	}
}
