package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;

/**
 * A collection of utility methods for cost strategies.
 * 
 * @author William Hawkins
 *
 */
public class CostStrategyUtils {
	
	/**
	 * Returns the aircraft that takes off last in the given sequence,
	 * which does not need to be sorted by takeoff times.
	 * 
	 * @param sequence
	 * @return
	 */
	public static Aircraft lastTakeOff(ArrayList<Aircraft> sequence) {
		int lastTakeOffTime = -1;
		Aircraft lastTakeOff = null;
		for (int i = 0; i < sequence.size(); i ++) {
			if (sequence.get(i).getRunwayReadyTime() > lastTakeOffTime) {
				lastTakeOff = sequence.get(i);
				lastTakeOffTime = lastTakeOff.getRunwayReadyTime();
			}
		}
		return lastTakeOff;
	}
	
	/**
	 * Returns an ArrayList of clones of the aircraft in the given sequence.
	 * 
	 * @param sequence
	 * @return
	 */
	public static ArrayList<Aircraft> cloneSequence(ArrayList<Aircraft> sequence) {
		ArrayList<Aircraft> clones = new ArrayList<Aircraft>();
		for (Aircraft ac : sequence) {
			clones.add(ac.clone());
		}
		return clones;
	}
	
	/**
	 * Calculates the earliest that the current aircraft should be released from the
	 * spot without violating constrained position shifting for the given release
	 * sequence.
	 * 
	 * @param releaseSequence
	 * @param current
	 * @param cps
	 * @return
	 */
	public static int safePrevTimeBound(ArrayList<Aircraft> releaseSequence, Aircraft current, int cps) {
		
		// Initialize values to the last aircraft released in the sequence
		// so far, since it could be the first to have already been pushed
		// back the maximum number of places
		Aircraft pred = releaseSequence.get(releaseSequence.size()-1);
		int predBound = pred.getOptReleaseTime();
		boolean foundPred = false;
		
		for (int i = releaseSequence.size()-2; i >= -1; i --) {
			int pos = i + 2;
			
			// Break out of the loop at the point that makes pred point to
			// the earliest aircraft that current MUST follow to obey CPS
			if (current.getFcfsIndex() - pos > cps) {
				foundPred = true;
				break;
			}
			if (pos - pred.getFcfsIndex() == cps) {
				foundPred = true;
				break;
			}
			
			// Set the earliest valid predecessor bound time, which will be pred's
			// since the code got to this point in the loop
			predBound = pred.getOptReleaseTime();
			
			// Move to the next one back if the aircraft at position pos can
			// follow current in the optimized release sequence
			if (i < 0) {
				break;
			}
			pred = releaseSequence.get(i);
		}
		
		// If a required predecessor was not found, this means the current
		// aircraft can go anywhere in the sequence, so its early bound with
		// respect to other aircraft can just be zero (it will be maxed with
		// its own spot ready time anyway)
		if (!foundPred) {
			return 0;
		}
		
		// Determine what the predecessor bound should be
		if (current.getFcfsIndex() > pred.getFcfsIndex()) {
			// Since the spot release sequence sorts in ascending order by FCFS
			// sequence number when the release times are the same, if current's
			// FCFS index is greater than that of pred's, we can just set
			// current's optimal release time to the same as pred's since it will
			// be sorted after pred automatically.
			predBound = pred.getOptReleaseTime();
		} else {
			// Otherwise, go with the earlier time among the stored last valid bound
			// time and pred's opt release time plus one second (the stored pred bound
			// could be the same as pred's, but the aircraft it belongs to could be ok
			// with following current due to a higher FCFS index than pred)
			predBound = Math.min(predBound, pred.getOptReleaseTime() + 1);
		}
		
		return predBound;
	}
}
