package edu.mit.ll.group43.surfaceoptimization.dp.factory.netgen;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.TreeSet;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.analysis.Logger;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.edge.EdgeValueStrategy;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.node.CPSNetworkNodeFactory;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;

public abstract class CPSNetworkGenerator {

	protected int CPS;
	protected CPSNetworkNodeFactory cpsNetworkNodeFactory;
	protected EdgeValueStrategy edgeValueStrategy;
	
	public CPSNetworkGenerator(int cps, CPSNetworkNodeFactory cpsNetworkNodeFactory,
			EdgeValueStrategy edgeValueStrategy) {
		CPS = cps;
		this.cpsNetworkNodeFactory = cpsNetworkNodeFactory;
		this.edgeValueStrategy = edgeValueStrategy;
	}
	
	public abstract ArrayList<TreeSet<CPSNetworkNode>> generateCPSNetworkNodes(ArrayList<Aircraft> fcfs);
	
	/**
	 * Generates the CPS network on which the Dynamic Programming algorithm
	 * will run to find the shortest path and therefore optimal sequence.
	 * 
	 * @param fcfs
	 */
	public ArrayList<TreeSet<CPSNetworkNode>> generateNetwork(ArrayList<Aircraft> fcfs) {
		
		Date start = new Date();
		ArrayList<TreeSet<CPSNetworkNode>> cpsStages = generateCPSNetworkNodes(fcfs);
		cpsStages = generateCPSNetworkNodeRelationships(cpsStages);
		Date end = new Date();
		if (Logger.timing) {
			Logger.buildNetwork(end.getTime()-start.getTime());
		}
		
		start = new Date();
		cpsStages = pruneNetwork(cpsStages);
		end = new Date();
		if (Logger.timing) {
			Logger.pruneNetwork(end.getTime()-start.getTime());
		}
		
		return cpsStages;
	}
	
	/**
	 * Creates a HashMap of valid locations for indexes of a FCFS sequence, with
	 * respect to CPS. The keys are the indexes of the new sequence, and the
	 * values are lists of indexes from the original sequence that can be placed
	 * at that position in the new sequence. Both sets of indexes are 1-based.
	 * 
	 * @param seqLength
	 * @return
	 */
	public HashMap<Integer, ArrayList<Integer>> getValidLocs(int seqLength) {
		HashMap<Integer, ArrayList<Integer>> validLocs = new HashMap<Integer, ArrayList<Integer>>();
		for (int i = 1; i <= seqLength; i ++) {
			validLocs.put(i, new ArrayList<Integer>());
			int minPos = Math.max(i - CPS, 1);
			int maxPos = Math.min(i + CPS, seqLength);
			for (int j = minPos; j <= maxPos; j ++) {
				validLocs.get(i).add(j);
			}
		}
		return validLocs;
	}
	
	/**
	 * Adds valid relationships (edges) between the nodes in the
	 * given CPS network.
	 * 
	 * @param network
	 * @return
	 */
	public ArrayList<TreeSet<CPSNetworkNode>> generateCPSNetworkNodeRelationships
		(ArrayList<TreeSet<CPSNetworkNode>> network) {
		
		// Connect the nodes with weighted edges
		for (int i = 0; i < network.size()-1; i ++) {
			
			// Get a reference to the next stage
			TreeSet<CPSNetworkNode> nextStage = network.get(i + 1);
			
			// Go through every node in this stage
			for (CPSNetworkNode node : network.get(i)) {
				
				// For every node in this stage, go through every node in the next stage
				for (CPSNetworkNode other : nextStage) {
					
					// Get the parts of the subsequences that need to match in order for
					// there to be an edge between 'node' and 'other'. 'begin' (inclusive)
					// and 'end' (exclusive) refer to the starting and ending points of node's
					// list to be compared, and can also be used to determine what to compare
					// from other's list of aircraft.
					int begin = Math.max(0, node.getAircraft().size() - 2 * CPS);
					int end = node.getAircraft().size();
					List<Aircraft> prevToMatch = node.getAircraft().subList(begin, end);
					List<Aircraft> nextToMatch = other.getAircraft().subList(0, end - begin);
					
					// If these two nodes are a match, create a relationship between the two
					if (prevToMatch.equals(nextToMatch)) {
						edgeValueStrategy.insertEdge(node, other);
					}
				}
			}
		}
		
		return network;
	}
	
	/**
	 * Marks all nodes that can't be visited from either (or both) of the
	 * starting or ending nodes.
	 * 
	 * @param network
	 * @return
	 */
	public ArrayList<TreeSet<CPSNetworkNode>> markDeadEnds(ArrayList<TreeSet<CPSNetworkNode>> network) {
		
		// Visit forwards
		ArrayList<CPSNetworkNode> current = new ArrayList<CPSNetworkNode>(network.get(0));
		while (current.size() > 0) {
			TreeSet<CPSNetworkNode> next = new TreeSet<CPSNetworkNode>();
			for (CPSNetworkNode node : current) {
				
				// Mark this node as being visited
				node.mark();
				
				// Add to the list for the next round of visits
				for (NetworkEdge nextEdge : node.getNextNodes()) {
					next.add((CPSNetworkNode) nextEdge.getTarget());
				}
			}
			current = new ArrayList<CPSNetworkNode>(next);
		}
		
		// Visit backwards
		current = new ArrayList<CPSNetworkNode>(network.get(network.size()-1));
		while (current.size() > 0) {
			TreeSet<CPSNetworkNode> prev = new TreeSet<CPSNetworkNode>();
			for (CPSNetworkNode node : current) {
				
				// If this node was already marked as being visited from the
				// starting node, unmark it so it isn't removed. Otherwise,
				// mark it for removal.
				if (node.isMarked()) {
					node.unmark();
				} else {
					node.mark();
				}
				
				// Add to the list for the next round of visits
				for (NetworkEdge prevEdge : node.getPrevNodes()) {
					prev.add((CPSNetworkNode) prevEdge.getSource());
				}
			}
			current = new ArrayList<CPSNetworkNode>(prev);
		}
		
		// Return network with dead-end nodes marked
		return network;
	}
	
	/**
	 * Takes in a marked network and removes all nodes and associated edges that
	 * have been marked for removal.
	 * 
	 * @param network
	 * @return
	 */
	public ArrayList<TreeSet<CPSNetworkNode>> prune(ArrayList<TreeSet<CPSNetworkNode>> network) {
		for (TreeSet<CPSNetworkNode> stage : network) {
			ArrayList<CPSNetworkNode> marked = new ArrayList<CPSNetworkNode>();
			for (CPSNetworkNode node : stage) {
				if (node.isMarked()) {
					node.removeRelationships();
					marked.add(node);
				}
			}
			stage.removeAll(marked);
		}
		
		return network;
	}
	
	/**
	 * Applies marking and pruning operations to the given network, then returns it.
	 * 
	 * @param network
	 * @return
	 */
	public ArrayList<TreeSet<CPSNetworkNode>> pruneNetwork(ArrayList<TreeSet<CPSNetworkNode>> network) {
		network = markDeadEnds(network);
		network = prune(network);
		return network;
	}
}
