package edu.mit.ll.group43.surfaceoptimization.measurement;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Arrays;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.AircraftAlphaComparator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.DefaultDepartureMinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator.ActionType;

/**
 * A collection of utility methods for classes that calculate metrics.
 * 
 * @author William Hawkins
 *
 */
public class MeasurementUtils {
	
	/**
	 * Returns the given sequence of aircraft in order of when
	 * they will reach the runway based on their spot release
	 * times and unimpeded taxi times.
	 * 
	 * @param sequence
	 * @return
	 */
	public static ArrayList<Aircraft> takeoffSequence(ArrayList<Aircraft> sequence) {

		Aircraft[] takeoffSequence = new Aircraft[sequence.size()];
		takeoffSequence = sequence.toArray(takeoffSequence);
		Arrays.sort(takeoffSequence, AircraftAlphaComparator.getInstance());
		
		ArrayList<Aircraft> takeoffList = new ArrayList<Aircraft>();
		takeoffList.addAll(Arrays.asList(takeoffSequence));
		return takeoffList;
	}
	
	/**
	 * Calculates what the takeoff times will be for the given sequence
	 * with respect to minimum separation times being preserved, in order
	 * of takeoff. Sequence should be sorted according to runway ready
	 * time before being passed to this method.
	 * 
	 * @param sequence
	 * @return
	 * @throws FileNotFoundException
	 */
	public static int[] takeoffTimes(ArrayList<Aircraft> sequence) throws FileNotFoundException {
		
		// Calculate the takeoff times
		DefaultDepartureMinimumSeparationCalculator calc = DefaultDepartureMinimumSeparationCalculator.getInstance();
		int[] takeoffs = new int[sequence.size()];
		takeoffs[0] = sequence.get(0).getRunwayReadyTime();
		
		for (int x = 1; x < takeoffs.length; x ++) {
			// Get the separation time between this aircraft and the one that
			// will take off before it
			WeightClass leader = sequence.get(x-1).getWeightClass();
			WeightClass follower = sequence.get(x).getWeightClass();
			int sep = calc.getSeparation(leader, ActionType.DEPARTURE, follower, ActionType.DEPARTURE);
			
			// The takeoff time for this aircraft is whichever happens last:
			// - When it is first permitted to go with respect to separation times
			// - When it is first ready to go (its alpha time)
			takeoffs[x] = Math.max(takeoffs[x-1] + sep, sequence.get(x).getRunwayReadyTime());
		}
		
		return takeoffs;
	}
}
