package edu.mit.ll.group43.surfaceoptimization.network.dp;
import java.util.ArrayList;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkNode;

/**
 * Represents a node in the CPS sequence optimization network.
 * 
 * @author William Hawkins
 *
 */
public class CPSNetworkNode implements NetworkNode, Comparable<CPSNetworkNode> {

	private static int created = 0;
	
	private ArrayList<NetworkEdge> prevNodes;
	private ArrayList<NetworkEdge> nextNodes;
	private Subsequence<Aircraft> aircraft;
	
	private NetworkEdge optPrevNode;
	private int optDistance;

	private boolean marked;

	/**
	 * Constructs a CPS network node from the given sequence
	 * of aircraft.
	 * 
	 * @param aircraft
	 */
	public CPSNetworkNode(Subsequence<Aircraft> aircraft) {
		this.aircraft = aircraft;
		this.prevNodes = new ArrayList<NetworkEdge>();
		this.nextNodes = new ArrayList<NetworkEdge>();
		this.marked = false;
		
		created ++;
	}
	
	/**
	 * Returns how many times the constructor of this class has been invoked.
	 * 
	 * @return
	 */
	public static int getCreated() {
		return created;
	}

	@Override
	public ArrayList<NetworkEdge> getPrevNodes() {
		return prevNodes;
	}

	@Override
	public void setPrevNodes(ArrayList<NetworkEdge> prevNodes) {
		this.prevNodes = prevNodes;
	}
	
	/**
	 * Add a relationship to the given node where this node
	 * is pointed to by it.
	 * 
	 * @param prevNode
	 */
	public void addPrevNode(NetworkEdge prevNode) {
		this.prevNodes.add(prevNode);
	}
	
	/**
	 * Remove the relationship between this node and the
	 * given node that points to it.
	 * 
	 * @param prevNode
	 */
	public void removePrevNode(NetworkEdge prevNode) {
		this.prevNodes.remove(prevNode);
	}

	@Override
	public ArrayList<NetworkEdge> getNextNodes() {
		return nextNodes;
	}

	@Override
	public void setNextNodes(ArrayList<NetworkEdge> nextNodes) {
		this.nextNodes = nextNodes;
	}
	 
	/**
	 * Add a relationship to the given node where this node
	 * points to it.
	 * 
	 * @param nextNode
	 */
	public void addNextNode(NetworkEdge nextNode) {
		this.nextNodes.add(nextNode);
	}
	
	/**
	 * Remove the relationship between this node and the
	 * given node to which it points.
	 * 
	 * @param nextNode
	 */
	public void removeNextNode(NetworkEdge nextNode) {
		this.nextNodes.remove(nextNode);
	}
	
	/**
	 * Remove relationships to this node from all nodes that point to
	 * and are pointed to by this node, then remove those relationships
	 * from this node as well.
	 * 
	 */
	public void removeRelationships() {
		for (NetworkEdge edge : this.prevNodes) {
			((CPSNetworkNode) edge.getSource()).removeNextNode(edge);
		}
		
		for (NetworkEdge edge : this.nextNodes) {
			((CPSNetworkNode) edge.getSource()).removePrevNode(edge);
		}
		
		this.prevNodes.clear();
		this.nextNodes.clear();
	}

	/**
	 * Get this node's predecessor node that gives it its optimal
	 * distance from the start node.
	 * 
	 * @return
	 */
	public NetworkEdge getOptPrevNode() {
		return optPrevNode;
	}

	/**
	 * Set this node's predecessor node that gives it its optimal
	 * distance from the start node.
	 * 
	 * @param optPrevNode
	 */
	public void setOptPrevNode(NetworkEdge optPrevNode) {
		this.optPrevNode = optPrevNode;
	}
	
	/**
	 * Get the optimal distance this node is from the
	 * start node.
	 * 
	 * @return
	 */
	public int getOptDistance() {
		return optDistance;
	}

	/**
	 * Set the optimal distance this node is from the
	 * start node.
	 * 
	 * @param optDistance
	 */
	public void setOptDistance(int optDistance) {
		this.optDistance = optDistance;
	}

	/**
	 * Get the subsequence of Aircraft this node represents
	 * 
	 * @return
	 */
	public Subsequence<Aircraft> getAircraft() {
		return aircraft;
	}
	
	public Aircraft getLastAircraft() {
		return aircraft.last();
	}
	
	public void mark() {
		this.marked = true;
	}
	
	public void unmark() {
		this.marked = false;
	}
	
	public boolean isMarked() {
		return this.marked;
	}
	
	@Override
	public int compareTo(CPSNetworkNode other) {
		if (this.equals(other)) {
			return 0;
		}
		return this.aircraft.compareTo(other.getAircraft());
	}
	
	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		} else if (other == null) {
			return false;
		}
		return this.aircraft.equals(((CPSNetworkNode) other).getAircraft());
	}
	
	@Override
	public int hashCode() {
		int hash = 1;
		hash = hash * 31 + aircraft.hashCode();
		return hash;
	}
}
