package edu.mit.ll.group43.surfaceoptimization.dp.factory.netgen;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.TreeSet;

import junit.framework.TestCase;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.DefaultDepartureMinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.edge.EdgeValueStrategy;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.edge.ZeroEdgeValueStrategy;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.node.CPSNetworkNodeFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.node.DefaultCPSNetworkNodeFactory;
import edu.mit.ll.group43.surfaceoptimization.input.Input;
import edu.mit.ll.group43.surfaceoptimization.input.InvalidFCFSReaderTypeException;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.network.dp.Subsequence;
import edu.mit.ll.group43.surfaceoptimization.network.dp.SubsequenceGeneratorNode;

public class CPSNetworkGeneratorTest extends TestCase {

	private MinimumSeparationCalculator<? extends WeightClass> calc;
	private CPSNetworkNodeFactory nodeFactory = new DefaultCPSNetworkNodeFactory();
	private EdgeValueStrategy edgeStrategy = new ZeroEdgeValueStrategy(calc);
	private String FCFS_FILE = "testdata/dpfcfs";
	private ArrayList<Aircraft> sequence;
	
	@BeforeClass
	public void initialize() throws InvalidFCFSReaderTypeException {
		try {
			calc = DefaultDepartureMinimumSeparationCalculator.getInstance();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * For some reason, initializing sequence in initialize() doesn't work.
	 * 
	 */
	@Before
	public void setUp() {
		try {
			sequence = (new Input()).generateFcfsSequence("text", FCFS_FILE);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (InvalidFCFSReaderTypeException e) {
			e.printStackTrace();
		}
	}
	
	private void assertValidLocs(CPSNetworkGenerator gen, int cps) throws FileNotFoundException, InvalidFCFSReaderTypeException {
		
		int size = sequence.size();
		HashMap<Integer, ArrayList<Integer>> validLocs = gen.getValidLocs(size);
		
		// Test that all locations given are valid
		for (int stage : validLocs.keySet()) {
			for (int index : validLocs.get(stage)) {
				assertTrue("Aircraft " + index + " can't be placed at position " + stage + " with CPS=" + cps, 
						Math.abs(stage - index) <= cps);
			}
		}
		
		// Test that all valid locations are given
		for (int stage : validLocs.keySet()) {
			// The first k stages should have k + (stage #) aircraft
			// The last k stages should have values that mirror those of the first k-1 stages
			// All stages in the middle (k < stage # <= n-k) should have 2k + 1 aircraft
			int correctNumPositions = Math.min(stage + cps, Math.min(2 * cps + 1, validLocs.size() - stage + cps + 1));
			assertEquals(correctNumPositions, validLocs.get(stage).size());
		}
	}
	
	@Test
	public void testGetValidLocs() throws FileNotFoundException, InvalidFCFSReaderTypeException {
		
		// Test the generator network CPS network generator
		CPSNetworkGenerator gen = new GeneratorNetworkCPSNetworkGenerator(1, nodeFactory, edgeStrategy);
		assertValidLocs(gen, 1);
		gen = new GeneratorNetworkCPSNetworkGenerator(2, nodeFactory, edgeStrategy);
		assertValidLocs(gen, 2);
		
		// Test the iterative CPS network generator
		gen = new IterativeCPSNetworkGenerator(1, nodeFactory, edgeStrategy);
		assertValidLocs(gen, 1);
		gen = new IterativeCPSNetworkGenerator(2, nodeFactory, edgeStrategy);
		assertValidLocs(gen, 2);
	}
	
	private void testGenerateSubsequenceGeneratorNetworkForCPS(int cps) throws FileNotFoundException, InvalidFCFSReaderTypeException {
		
		CPSNetworkGenerator gen = new GeneratorNetworkCPSNetworkGenerator(cps, nodeFactory, edgeStrategy);
		
		int size = sequence.size();
		
		HashMap<Integer, ArrayList<Integer>> validLocs = gen.getValidLocs(size);
		ArrayList<ArrayList<SubsequenceGeneratorNode>> network 
			= ((GeneratorNetworkCPSNetworkGenerator) gen).generateSubsequenceGeneratorNetwork(size);
		
		// Make sure the two have the same number of stages
		assertEquals(validLocs.size(), network.size());
		
		// Make sure the stages have the same number of members
		for (int i = 0; i < validLocs.size(); i ++) {
			assertEquals(validLocs.get(i+1).size(), network.get(i).size());
		}
		
		// Make sure there are edges between every pair of nodes that are
		// in neighboring stages
		for (int i = 1; i < network.size(); i ++) {
			ArrayList<SubsequenceGeneratorNode> stage = network.get(i);
			ArrayList<SubsequenceGeneratorNode> prevStage = network.get(i-1);
			
			// Each node in this stage should have the same number of
			// previous nodes as there are in the previous stage
			for (SubsequenceGeneratorNode node : stage) {
				TreeSet<Integer> nodes = new TreeSet<Integer>();
				for (NetworkEdge edge : node.getPrevNodes()) {
					nodes.add(((SubsequenceGeneratorNode) edge.getSource()).getFcfs());
				}
				assertEquals(nodes.size(), prevStage.size());
			}
			
			// Each node in the previous stage should have the same number
			// of next nodes as there are in this stage
			for (SubsequenceGeneratorNode node : prevStage) {
				TreeSet<Integer> nodes = new TreeSet<Integer>();
				for (NetworkEdge edge : node.getNextNodes()) {
					nodes.add(((SubsequenceGeneratorNode) edge.getTarget()).getFcfs());
				}
				assertEquals(nodes.size(), stage.size());
			}
		}
	}
	
	@Test
	public void testGenerateSubsequenceGeneratorNetwork() throws FileNotFoundException, InvalidFCFSReaderTypeException {
		testGenerateSubsequenceGeneratorNetworkForCPS(1);
		testGenerateSubsequenceGeneratorNetworkForCPS(2);
	}
	
	private <T extends Comparable<T>> void noRepeats(Subsequence<T> path) {
		TreeSet<T> indexSet = new TreeSet<T>();
		for (T elem : path) {
			assertTrue(indexSet.add(elem));
		}
	}
	
	private void testGenerateValidPathsForCPS(int cps, int aircraft) {

		CPSNetworkGenerator gen = new GeneratorNetworkCPSNetworkGenerator(cps, nodeFactory, edgeStrategy);
		
		ArrayList<ArrayList<SubsequenceGeneratorNode>> network 
			= ((GeneratorNetworkCPSNetworkGenerator) gen).generateSubsequenceGeneratorNetwork(aircraft);
		network = ((GeneratorNetworkCPSNetworkGenerator) gen).generateValidPaths(network);
		
		for (ArrayList<SubsequenceGeneratorNode> stage : network) {
			for (SubsequenceGeneratorNode node : stage) {
				for (Subsequence<Integer> path : node.getValidPaths()) {
					
					// No indexes should be repeated
					noRepeats(path);
					
					// Last index of the path must be that of this node
					assertTrue("Last index in path of node " + node.getFcfs() + " in stage " +
							network.indexOf(stage) + " was " + path.last(),
							path.last().equals(node.getFcfs()));
					
					// Length of the path should be the minimum of the stage number and
					// 2k + 1
					int stageNum = network.indexOf(stage);
					int seqLength = Math.min(stageNum + 1, 2 * cps + 1);
					assertEquals("Path in stage " + (stageNum + 1) + " was " +
							path.size() + " instead of the expected " + seqLength, path.size(), seqLength);
					
					//TODO: Test that all valid paths are present?
				}
			}
		}
	}
	
	@Test
	public void testGenerateValidPaths() throws FileNotFoundException {
		
		testGenerateValidPathsForCPS(1, 5);
		testGenerateValidPathsForCPS(1, 10);
		testGenerateValidPathsForCPS(1, 50);
		
		testGenerateValidPathsForCPS(2, 5);
		testGenerateValidPathsForCPS(2, 10);
		testGenerateValidPathsForCPS(2, 50);
	}
	
	private void validAircraftOrder(Subsequence<Aircraft> aircraft, int stage, int cps) {
		int size = aircraft.size();
		for (Aircraft a : aircraft) {
			int pos = stage - size + aircraft.indexOf(a) + 1;
			assertTrue(Math.abs(pos - a.getFcfsIndex()) <= cps);
		}
	}
	
	@Test
	public void testGenerateNetwork() throws FileNotFoundException, InvalidFCFSReaderTypeException {
		
		//TODO Extend this testing for CPS-2
		
		//TODO Extend this testing to iterative CPS network generator
		
		CPSNetworkGenerator gen = new GeneratorNetworkCPSNetworkGenerator(1, nodeFactory, edgeStrategy);
		
		ArrayList<ArrayList<SubsequenceGeneratorNode>> network 
			= ((GeneratorNetworkCPSNetworkGenerator) gen).generateSubsequenceGeneratorNetwork(5);
		network = ((GeneratorNetworkCPSNetworkGenerator) gen).generateValidPaths(network);
		ArrayList<TreeSet<CPSNetworkNode>> cpsNetwork = gen.generateCPSNetworkNodes(sequence);
		cpsNetwork = gen.generateCPSNetworkNodeRelationships(cpsNetwork);
		
		// The two networks must have the same number of stages
		assertEquals(network.size(), cpsNetwork.size());
		
		for (int i = 0; i < network.size(); i ++) {
			// Corresponding stages from the two networks must have the same
			// number of sequences
			ArrayList<Subsequence<Integer>> seqs = new ArrayList<Subsequence<Integer>>();
			for (SubsequenceGeneratorNode node : network.get(i)) {
				seqs.addAll(node.getValidPaths());
			}
			assertEquals("There were " + seqs.size() + " index sequences, but there were " + cpsNetwork.get(i).size() +
					" aircraft sequences in stage " + (i+1),
					seqs.size(), cpsNetwork.get(i).size());
			
			for (CPSNetworkNode node : cpsNetwork.get(i)) {
				// All nodes in a stage must have the same number of aircraft (Min(stage, 2k+1))
				assertEquals(node.getAircraft().size(), Math.min(i+1, 3));
				
				// There must be no repeats in any node
				noRepeats(node.getAircraft());
				
				// All aircraft must be able to be in their respective positions within a node
				validAircraftOrder(node.getAircraft(), i+1, 1);
				
				// All relationships must be valid
				for (NetworkEdge nextEdge : node.getNextNodes()) {
					int begin = Math.max(0, node.getAircraft().size() - 2);
					int end = node.getAircraft().size();
					List<Aircraft> prevToMatch = node.getAircraft().subList(begin, end);
					List<Aircraft> nextToMatch = ((CPSNetworkNode) nextEdge.getTarget()).getAircraft().subList(0, end - begin);
					assertTrue(prevToMatch.equals(nextToMatch));
				}
			}
		}
		
		//TODO: ALL valid relationships must be present
	}
	
	@Test
	public void testMarkDeadEnds() throws FileNotFoundException, InvalidFCFSReaderTypeException {
		
		//TODO: Extend to CPS-2
		
		//TODO: Extend to iterative method by instead counting how many are NOT marked
		
		CPSNetworkGenerator gen = new GeneratorNetworkCPSNetworkGenerator(1, nodeFactory, edgeStrategy);
		
		ArrayList<TreeSet<CPSNetworkNode>> cpsNetwork = gen.generateCPSNetworkNodes(sequence);
		cpsNetwork = gen.generateCPSNetworkNodeRelationships(cpsNetwork);
		
		cpsNetwork = gen.markDeadEnds(cpsNetwork);
		
		int[] expNumMarked = {0, 1, 2, 3, 0};
		
		for (int i = 0; i < cpsNetwork.size(); i ++) {
			int marked = 0;
			for (CPSNetworkNode node : cpsNetwork.get(i)) {
				if (node.isMarked()) {
					marked ++;
				}
			}
			assertEquals(marked, expNumMarked[i]);
		}
	}
	
	@Test
	public void testPrune() throws FileNotFoundException, InvalidFCFSReaderTypeException {
		
		//TODO Extend to CPS-2
		
		//TODO Extend to iterative method
		
		CPSNetworkGenerator gen = new GeneratorNetworkCPSNetworkGenerator(1, nodeFactory, edgeStrategy);
		
		ArrayList<TreeSet<CPSNetworkNode>> cpsNetwork = gen.generateCPSNetworkNodes(sequence);
		cpsNetwork = gen.generateCPSNetworkNodeRelationships(cpsNetwork);
		
		cpsNetwork = gen.markDeadEnds(cpsNetwork);
		cpsNetwork = gen.prune(cpsNetwork);
		
		// Test that there are no marked nodes left
		for (TreeSet<CPSNetworkNode> stage : cpsNetwork) {
			for (CPSNetworkNode node : stage) {
				assertFalse(node.isMarked());
			}
		}
		
		// Test that there are no edges to or from marked nodes left, and that
		// all nodes that should have them have edges to or from at least
		// one node
		for (TreeSet<CPSNetworkNode> stage : cpsNetwork) {
			int stageNum = cpsNetwork.indexOf(stage) + 1;
			for (CPSNetworkNode node : stage) {
				// Test prev nodes
				if (stageNum > 1) {
					assertTrue(node.getPrevNodes().size() > 0);
					for (NetworkEdge edge : node.getPrevNodes()) {
						assertFalse(((CPSNetworkNode) edge.getSource()).isMarked());
					}
				} else {
					assertEquals(0, node.getPrevNodes().size());
				}
				
				// Test next nodes
				if (stageNum < cpsNetwork.size()) {
					assertTrue(node.getNextNodes().size() > 0);
					for (NetworkEdge edge : node.getNextNodes()) {
						assertFalse(((CPSNetworkNode) edge.getTarget()).isMarked());
					}
				} else {
					assertEquals(0, node.getNextNodes().size());
				}
			}
		}
	}
	
	//TODO: Test solving the shortest path / optimal schedule
}
